"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ArrayOfShape = exports.ObjectOfShape = exports.ArrayShape = exports.ObjectShape = exports.set = exports.TYPE_ARRAY = exports.TYPE_OBJECT = void 0;

var _util = require("./util");

var _processor = require("./processor");

var _context = require("./context");

var _error = require("./error");

var _future = require("./future");

var _inspect = require("./inspect");

const TYPE_OBJECT = 'TYPE_OBJECT';
exports.TYPE_OBJECT = TYPE_OBJECT;
const TYPE_ARRAY = 'TYPE_ARRAY';
exports.TYPE_ARRAY = TYPE_ARRAY;

const set = (object, key, value) => {
  const setters = object[_future.SHAPE];
  const setter = setters && setters[key];
  const context = new _context.Context(value, key, object, [key]);
  context.parent = object;

  if (!setter) {
    throw (0, _error.error)('SHAPE_NOT_FOUND');
  }

  return setter(value, false, context);
};

exports.set = set;

function clean(context) {
  context.rawParent = context.parent = null;
}

const isObject = v => Object(v) === v;

function reducer(prev, factory) {
  return factory.call(this, prev);
}

class Shape {
  constructor(shape, clean) {
    this._shape = shape;
    this._clean = clean;
  }

  from(args, context, options) {
    if (!isObject(context.input)) {
      const error = context.errorByCode('NOT_OBJECT', context.key);
      return options.promise.reject(error);
    }

    const values = this._create(context);

    const setters = Object.create(null);
    (0, _util.defineValue)(values, _future.SHAPE, setters);
    (0, _inspect.attach)(values);

    const tasks = this._tasks(context).map(([context, skema]) => {
      context.parent = values;
      return () => new _processor.Processor({
        options,
        skema,
        args,
        context,
        values,
        setters
      }).process().then(() => clean(context), error => {
        clean(context);
        return options.promise.reject(error);
      });
    });

    return options.promiseExtra.series(tasks, reducer).then(() => values);
  }

}

class ObjectShape extends Shape {
  _create(context) {
    return this._clean ? Object.create(null) : (0, _util.simpleClone)(context.input);
  }

  _tasks(context) {
    const shape = this._shape;
    return Object.keys(shape).map(key => [context.descend(key), shape[key]]);
  }

}

exports.ObjectShape = ObjectShape;

class ArrayShape extends Shape {
  _create(context) {
    return this._clean ? [] : [].concat(context.input);
  }

  _tasks(context) {
    return this._shape.map((v, i) => [context.descend(i), v]);
  }

}

exports.ArrayShape = ArrayShape;

class ObjectOfShape extends Shape {
  _create() {
    return Object.create(null);
  }

  _tasks(context) {
    return Object.keys(context.input).map(key => [context.descend(key), this._shape]);
  }

}

exports.ObjectOfShape = ObjectOfShape;

class ArrayOfShape extends Shape {
  _create() {
    return [];
  }

  _tasks(context) {
    const tasks = [];
    const {
      input
    } = context;
    const {
      length
    } = input;
    let i = 0; // Iterate every array item

    for (; i < length; i++) {
      tasks.push([context.descend(i), this._shape]);
    }

    return tasks;
  }

}

exports.ArrayOfShape = ArrayOfShape;